@@ -2,4 +2,4 @@ import sys
 from .cli import main
 
 if __name__ == "__main__":
-    main(sys.argv)
+    main(sys.argv[1:])
