@@ -1,4 +1,4 @@
-from typing import Optional, Any, List, Dict, ForwardRef
+from typing import Optional, Any, List, Dict, ForwardRef, Union
 
 import re
 import builtins
@@ -8,7 +8,7 @@ import uuid
 from pydantic import BaseModel, Field, root_validator, Extra
 
 from .json import JSONPointer
-from .errors import ReferenceResolutionError, SpecError
+from .errors import ReferenceResolutionError, SpecError, OperationParameterValidationError
 
 # from . import me
 
@@ -76,7 +76,7 @@ class PathsBase(ObjectBase):
 
 class RootBase:
     @staticmethod
-    def resolve(api, root, obj, _PathItem, _Reference):
+    def resolve(api: "OpenAPI", root: "RootBase", obj, _PathItem, _Reference):
         from . import v20, v30, v31
 
         def replaceSchemaReference(data):
@@ -237,9 +237,11 @@ class SchemaBase:
         """
         r = BaseModel.__getstate__(self)
         try:
-            if "_model_type" in r["__private_attribute_values__"]:
-                r["__private_attribute_values__"] = r["__private_attribute_values__"].copy()
-                del r["__private_attribute_values__"]["_model_type"]
+            for i in ["_model_type", "_model_types"]:
+                if i in r["__private_attribute_values__"]:
+                    r["__private_attribute_values__"] = r["__private_attribute_values__"].copy()
+                    del r["__private_attribute_values__"][i]
+
         except Exception:
             pass
         return r
@@ -271,14 +273,19 @@ class SchemaBase:
 
     def set_type(
         self, names: List[str] = None, discriminators: List[DiscriminatorBase] = None, extra: "SchemaBase" = None
-    ):
+    ) -> BaseModel:
         from .model import Model
 
+        if not hasattr(self, "_model_types"):
+            self._model_types = list()
+
         if extra is None:
             self._model_type = Model.from_schema(self, names, discriminators)
             return self._model_type
         else:
-            return Model.from_schema(self, names, discriminators, extra)
+            r = Model.from_schema(self, names, discriminators, extra)
+            self._model_types.append(r)
+            return r
 
     def get_type(
         self,
@@ -286,7 +293,7 @@ class SchemaBase:
         discriminators: List[DiscriminatorBase] = None,
         extra: "SchemaBase" = None,
         fwdref: bool = False,
-    ):
+    ) -> Union[BaseModel, ForwardRef]:
         try:
             if extra is None:
                 return self._model_type
@@ -326,13 +333,13 @@ class SchemaBase:
 
 
 class OperationBase:
-    def _validate_path_parameters(self, pi: "PathItem", path):
+    def _validate_path_parameters(self, pi: "PathItem", path_, loc):
         """
         Ensures that all parameters for this path are valid
         """
-        assert isinstance(path, str)
+        assert isinstance(path_, str)
         # FIXME { and } are allowed in parameter name, regex can't handle this e.g. {name}}
-        path = frozenset(re.findall(r"{([a-zA-Z0-9\-\._~]+)}", path))
+        path = frozenset(re.findall(r"{([a-zA-Z0-9\-\._~]+)}", path_))
 
         op = frozenset(map(lambda x: x.name, filter(lambda c: c.in_ == "path", self.parameters)))
         pi = frozenset(map(lambda x: x.name, filter(lambda c: c.in_ == "path", pi.parameters)))
@@ -342,16 +349,20 @@ class OperationBase:
             # FIXME
             #   OpenAPI does not allow RFC 6570 URI templates
             #   but name:\d+ may be valid though
-            raise SpecError(f"Parameter names are invalid: {invalid}")
+            raise OperationParameterValidationError(path_, *loc, f"Parameter names are invalid: {invalid}")
 
         r = (op | pi) - path
         if r:
-            raise SpecError(f"Parameter name{'s' if len(r) > 1 else ''} not found in path: {', '.join(sorted(r))}")
+            raise OperationParameterValidationError(
+                path_, *loc, f"Parameter name{'s' if len(r) > 1 else ''} not found in path: {', '.join(sorted(r))}"
+            )
 
         r = path - (op | pi)
         if r:
-            raise SpecError(
-                f"Parameter name{'s' if len(r) > 1 else ''} not found in parameters: {', '.join(sorted(r))}"
+            raise OperationParameterValidationError(
+                path_,
+                *loc,
+                f"Parameter name{'s' if len(r) > 1 else ''} not found in parameters: {', '.join(sorted(r))}",
             )
 
 
