@@ -1,3 +1,4 @@
+import copy
 import sys
 
 if sys.version_info >= (3, 9):
@@ -5,15 +6,19 @@ if sys.version_info >= (3, 9):
 else:
     import pathlib3x as pathlib
 
-from typing import List, Dict, Union, Callable, Tuple
+from typing import List, Dict, Set, Union, Callable, Tuple, Any
+import collections
 import inspect
 import logging
+import copy
+import pickle
 
 import httpx
 import yarl
 from pydantic import BaseModel
 
 from aiopenapi3.v30.general import Reference
+import aiopenapi3.request
 from .json import JSONReference
 from . import v20
 from . import v30
@@ -56,19 +61,38 @@ class OpenAPI:
         loader=None,
         plugins: List[Plugin] = None,
         use_operation_tags: bool = False,
-    ):
+    ) -> "OpenAPI":
+        """
+        Create a synchronous OpenAPI object from a description document.
+
+        :param url: the url of the description document
+        :param session_factory: used to create the session for http/s io
+        :param loader: the backend to access referenced description documents
+        :param plugins: potions to cure defects in the description document or requests/responses
+        :param use_operation_tags: honor tags
+        """
+
         resp = session_factory().get(url)
         return cls._load_response(url, resp, session_factory, loader, plugins, use_operation_tags)
 
     @classmethod
     async def load_async(
         cls,
-        url,
+        url: str,
         session_factory: Callable[[], httpx.AsyncClient] = httpx.AsyncClient,
-        loader=None,
+        loader: Loader = None,
         plugins: List[Plugin] = None,
         use_operation_tags: bool = False,
-    ):
+    ) -> "OpenAPI":
+        """
+        Create an asynchronous OpenAPI object from a description document.
+
+        :param url: the url of the description document
+        :param session_factory: used to create the session for http/s io
+        :param loader: the backend to access referenced description documents
+        :param plugins: potions to cure defects in the description document or requests/responses
+        :param use_operation_tags: honor tags
+        """
         async with session_factory() as client:
             resp = await client.get(url)
         return cls._load_response(url, resp, session_factory, loader, plugins, use_operation_tags)
@@ -82,13 +106,24 @@ class OpenAPI:
     @classmethod
     def load_file(
         cls,
-        url,
+        url: str,
         path: Union[str, pathlib.Path, yarl.URL],
-        session_factory: Callable[[], httpx.AsyncClient] = httpx.AsyncClient,
+        session_factory: Callable[[], Union[httpx.AsyncClient, httpx.Client]] = httpx.AsyncClient,
         loader: Loader = None,
         plugins: List[Plugin] = None,
         use_operation_tags: bool = False,
-    ):
+    ) -> "OpenAPI":
+        """
+        Create an OpenAPI object from a description document file.
+
+
+        :param url: the fictive url of the description document
+        :param path: description document location
+        :param session_factory: used to create the session for http/s io, defaults to use an AsyncClient
+        :param loader: the backend to access referenced description documents
+        :param plugins: potions to cure defects in the description document or requests/responses
+        :param use_operation_tags: honor tags
+        """
         assert loader
         if not isinstance(path, yarl.URL):
             path = yarl.URL(str(path))
@@ -99,37 +134,46 @@ class OpenAPI:
     def loads(
         cls,
         url: str,
-        data,
-        session_factory: Callable[[], httpx.AsyncClient] = httpx.AsyncClient,
+        data: str,
+        session_factory: Callable[[], Union[httpx.AsyncClient, httpx.Client]] = httpx.AsyncClient,
         loader=None,
         plugins: List[Plugin] = None,
         use_operation_tags: bool = False,
-    ):
+    ) -> "OpenAPI":
+        """
+
+        :param url: the url of the description document
+        :param data: description document
+        :param session_factory: used to create the session for http/s io, defaults to use an AsyncClient
+        :param loader: the backend to access referenced description documents
+        :param plugins: potions to cure defects in the description document or requests/responses
+        :param use_operation_tags: honor tags
+        """
         if loader is None:
             loader = NullLoader()
         data = loader.parse(Plugins(plugins or []), yarl.URL(url), data)
         return cls(url, data, session_factory, loader, plugins, use_operation_tags)
 
-    def _parse_obj(self, raw_document):
-        v = raw_document.get("openapi", None)
+    def _parse_obj(self, document: Dict[str, Any]) -> RootBase:
+        v = document.get("openapi", None)
         if v:
             v = list(map(int, v.split(".")))
             if v[0] == 3:
                 if v[1] == 0:
-                    return v30.Root.parse_obj(raw_document)
+                    return v30.Root.parse_obj(document)
                 elif v[1] == 1:
-                    return v31.Root.parse_obj(raw_document)
+                    return v31.Root.parse_obj(document)
                 else:
                     raise ValueError(f"openapi version 3.{v[1]} not supported")
             else:
                 raise ValueError(f"openapi major version {v[0]} not supported")
             return
 
-        v = raw_document.get("swagger", None)
+        v = document.get("swagger", None)
         if v:
             v = list(map(int, v.split(".")))
             if v[0] == 2 and v[1] == 0:
-                return v20.Root.parse_obj(raw_document)
+                return v20.Root.parse_obj(document)
             else:
                 raise ValueError(f"swagger version {'.'.join(v)} not supported")
         else:
@@ -137,25 +181,24 @@ class OpenAPI:
 
     def __init__(
         self,
-        url,
-        document,
+        url: str,
+        document: Dict[str, Any],
         session_factory: Callable[[], Union[httpx.Client, httpx.AsyncClient]] = httpx.AsyncClient,
-        loader=None,
+        loader: Loader = None,
         plugins: List[Plugin] = None,
         use_operation_tags: bool = True,
-    ):
+    ) -> "OpenAPI":
         """
         Creates a new OpenAPI document from a loaded spec file.  This is
         overridden here because we need to specify the path in the parent
         class' constructor.
 
+        :param url: the url of the description document
         :param document: The raw OpenAPI file loaded into python
-        :type document: dct
         :param session_factory: default uses new session for each call, supply your own if required otherwise.
-        :type session_factory: returns httpx.AsyncClient or http.Client
+        :param loader: the Loader for the description document(s)
+        :param plugins: list of plugins
         :param use_operation_tags: honor tags
-        :type use_operation_tags: bool
-
         """
 
         self._base_url: yarl.URL = yarl.URL(url)
@@ -191,11 +234,12 @@ class OpenAPI:
         log.init()
         self.log = logging.getLogger("aiopenapi3.OpenAPI")
 
-        document = self.plugins.document.parsed(url=self._base_url, document=document).document
-
+        """
+        Document Plugins get called via OpenAPI.load… - this is processed already
+        """
         self._root = self._parse_obj(document)
 
-        self._documents[yarl.URL(url)] = self._root
+        self._documents[self._base_url] = self._root
 
         self._init_session_factory(session_factory)
         self._init_references()
@@ -237,12 +281,19 @@ class OpenAPI:
         processed = set()
         while True:
             values = {id(x): x for x in self._documents.values()}
+            names = {id(v): k for k, v in self._documents.items()}
             todo = set(values.keys()) - processed
             if not todo:
                 break
             for i in todo:
-                values[i]._resolve_references(self)
+                #                print(names[i])
+                try:
+                    values[i]._resolve_references(self)
+                except ReferenceResolutionError as e:
+                    e.document = names[i]
+                    raise
             processed = set(values.keys())
+        return
 
     #        for i in self._documents.values():
     #            i._resolve_references(self)
@@ -253,7 +304,7 @@ class OpenAPI:
                 for path, obj in self.paths.items():
                     for m in obj.__fields_set__ & HTTP_METHODS:
                         op = getattr(obj, m)
-                        op._validate_path_parameters(obj, path)
+                        op._validate_path_parameters(obj, path, (m, op.operationId))
                         if op.operationId is None:
                             continue
                         for r, response in op.responses.items():
@@ -279,7 +330,7 @@ class OpenAPI:
                 for path, obj in self.paths.items():
                     for m in obj.__fields_set__ & HTTP_METHODS:
                         op = getattr(obj, m)
-                        op._validate_path_parameters(obj, path)
+                        op._validate_path_parameters(obj, path, (m, op.operationId))
                         if op.operationId is None:
                             continue
                         for r, response in op.responses.items():
@@ -299,19 +350,17 @@ class OpenAPI:
         self._operationindex = OperationIndex(self, use_operation_tags)
 
     @staticmethod
-    def _iterate_schemas(schemas, d, r):
-        if not d:
-            return r
+    def _iterate_schemas(schemas: Dict[int, SchemaBase], next: Set[int], processed: Set[int]):
+        if not next:
+            return processed
 
-        r.update(d)
-
-        import collections
+        processed.update(next)
 
         new = collections.ChainMap(
             *[
                 dict(
                     filter(
-                        lambda z: z[0] not in r,
+                        lambda z: z[0] not in processed,
                         map(
                             lambda y: (id(y._target), y._target) if isinstance(y, ReferenceBase) else (id(y), y),
                             filter(
@@ -336,19 +385,19 @@ class OpenAPI:
                         ),
                     )
                 )
-                for i in d
+                for i in next
             ]
         )
 
         sets = new.keys()
         schemas.update(new)
-        r.update(sets)
+        processed.update(sets)
 
-        return OpenAPI._iterate_schemas(schemas, sets, r)
+        return OpenAPI._iterate_schemas(schemas, sets, processed)
 
     def _init_schema_types(self):
-        byname = dict()
-        data = set()
+        byname: Dict[str, SchemaBase] = dict()
+        data: Set[int] = set()
         if isinstance(self._root, v20.Root):
             # Schema
             for byid in map(lambda x: x.definitions, self._documents.values()):
@@ -358,6 +407,10 @@ class OpenAPI:
             # Request
 
             # Response
+            for byid in map(lambda x: x.responses, self._documents.values()):
+                for name, response in filter(lambda v: isinstance(v[1].schema_, SchemaBase), byid.items()):
+                    byname[response.schema_._get_identity(name=name)] = response.schema_
+                    data.add(id(response.schema_))
 
         elif isinstance(self._root, (v30.Root, v31.Root)):
             # Schema
@@ -400,20 +453,25 @@ class OpenAPI:
                         byname[media_type.schema_._get_identity("R")] = media_type.schema_
                         data.add(id(media_type.schema_))
 
-        byid = {id(i): i for i in byname.values()}
-        data = set(byid.keys())
-        todo = self._iterate_schemas(byid, data, set())
+        byid: Dict[int, SchemaBase] = {id(i): i for i in byname.values()}
+        data: Set[int] = set(byid.keys())
+        todo: Set[int] = self._iterate_schemas(byid, data, set())
 
-        types = dict()
+        types: Dict[int, "BaseModel"] = dict()
         for i in todo | data:
             b = byid[i]
             types[b._get_identity("X")] = b.get_type()
 
         for name, schema in types.items():
             if not (inspect.isclass(schema) and issubclass(schema, BaseModel)):
+                # primitive types: str, int …
                 continue
             try:
                 schema.update_forward_refs(**types)
+                thes = byname.get(name, None)
+                if thes is not None:
+                    for v in byid[id(thes)]._model_types:
+                        v.update_forward_refs(**types)
             except Exception as e:
                 raise e
 
@@ -443,9 +501,10 @@ class OpenAPI:
         elif isinstance(self._root, (v30.Root, v31.Root)):
             return self._base_url.join(yarl.URL(self._root.servers[0].url))
 
-    # public methods
     def authenticate(self, *args, **kwargs):
         """
+        authenticate, multiple authentication schemes can be used simultaneously serving "or" or "and"
+        authentication schemes
 
         :param args: None to remove all credentials / reset the authorizations
         :param kwargs: scheme=value
@@ -474,16 +533,61 @@ class OpenAPI:
         return self._parse_obj(data)
 
     @property
-    def _(self):
+    def _(self) -> OperationIndex:
+        """
+        the sad smiley interface
+        access operations by operationId
+        """
         return self._operationindex
 
+    def createRequest(self, operationId: Union[str, Tuple[str, str]]) -> aiopenapi3.request.RequestBase:
+        """
+        create a Request
+
+        lookup the Operation by operationId or path,method
+
+        the type of Request returned depends on the session_factory of the OpenAPI object and OpenAPI/Swagger version
+
+        :param operationId: the operationId or tuple(path,method)
+        :return: the returned Request is either :class:`aiopenapi3.request.RequestBase` or -
+            in case of a httpx.AsyncClient session_factory - :class:`aiopenapi3.request.AsyncRequestBase`
+        """
+        if isinstance(operationId, str):
+            p = operationId.split(".")
+            req = self._operationindex
+            for i in p:
+                req = getattr(req, i)
+            assert isinstance(req, aiopenapi3.request.RequestBase)
+        else:
+            path, method = operationId
+            op = getattr(self._root.paths[path], method)
+            req = self._createRequest(self, method, path, op)
+        return req
+
     def resolve_jr(self, root: RootBase, obj, value: Reference):
+        """
+        Resolve a `JSON Reference<https://datatracker.ietf.org/doc/html/draft-pbryan-zyp-json-ref-03>`_ in our documents
+
+        :param root:
+        :param obj:
+        :param value:
+        :return:
+        """
         url, jp = JSONReference.split(value.ref)
         if url != "":
             url = yarl.URL(url)
             if url not in self._documents:
                 self.log.debug(f"Resolving {value.ref} - Description Document {url} unknown …")
-                self._documents[url] = self._load(url)
+                try:
+                    self._documents[url] = self._load(url)
+                except FileNotFoundError:
+                    err = ReferenceResolutionError(f"not found {url}")
+                    for k, v in self._documents.items():
+                        if v == root:
+                            err.document = k
+                            break
+                    raise err
+
             root = self._documents[url]
 
         try:
@@ -491,4 +595,73 @@ class OpenAPI:
         except ReferenceResolutionError as e:
             # add metadata to the error
             e.element = obj
+            for k, v in self._documents.items():
+                if v == root:
+                    e.document = k
+                    break
             raise
+
+    def __copy__(self) -> "OpenAPI":
+        """
+        shallow copy of an API object allows for a quick & low resource way to interface multiple
+        services using the same api instad of creating a new OpenAPI object from the description document for each
+
+        after setting the _base_url & .authenticate() it is ready to use
+        :return: aiopenapi3.OpenAPI
+        """
+        api = OpenAPI("/test.yaml", {"openapi": "3.0.0", "info": {"title": "", "version": ""}})
+        api._root = self._root
+        api.plugins = self.plugins
+        api._documents = self._documents
+        api._security = copy.deepcopy(self._security)
+        api._createRequest = self._createRequest
+        api._session_factory = self._session_factory
+        api.loader = self.loader
+        return api
+
+    def clone(self, baseurl: yarl.URL = None) -> "OpenAPI":
+        """
+        shallwo copy the api object
+        optional set a base url
+
+        :param baseurl:
+        """
+        api = copy.copy(self)
+        if baseurl:
+            api._base_url = baseurl
+        return api
+
+    @staticmethod
+    def cache_load(path: pathlib.Path, plugins: List[Plugin] = None, session_factory=None) -> "OpenAPI":
+        """
+        read a pickle api object from path and init the schema types
+
+        :param path: cache path
+        """
+        with path.open("rb") as f:
+            api = pickle.load(f)
+
+        api._init_schema_types()
+
+        if session_factory is not None:
+            api._session_factory = session_factory
+
+        if plugins is not None:
+            api._init_plugins(plugins)
+            api.plugins.init.initialized(initialized=api._root)
+
+        return api
+
+    def cache_store(self, path: pathlib.Path) -> None:
+        """
+        write the pickled api object to Path
+        to dismiss potentially local defined objects loader, plugins and the session_factory are dropped
+
+        :param path: cache path
+        """
+
+        restore = (self.loader, self.plugins, self._session_factory)
+        self.loader = self._session_factory = self.plugins = None
+        with path.open("wb") as f:
+            pickle.dump(self, f)
+        self.loader, self.plugins, self._session_factory = restore
