@@ -1,8 +1,10 @@
+import abc
 import collections
 import io
 from contextlib import closing
-from typing import Dict, Tuple, Union, Any, Optional
+from typing import Dict, Tuple, Union, Any, Optional, List
 
+import httpx
 import yarl
 
 try:
@@ -20,7 +22,7 @@ except:  # <= Python 3.10
 
 from .base import HTTP_METHODS
 from .version import __version__
-from .errors import SpecError
+from .errors import SpecError, RequestError
 
 
 class RequestParameter:
@@ -46,15 +48,22 @@ class RequestBase:
         self.operation = operation
         self.req: RequestParameter = RequestParameter(self.path)
 
-    def __call__(self, *args, **kwargs):
-        return self.request(*args, **kwargs)
+    def __call__(self, *args, return_headers: bool = False, **kwargs) -> Union[Any, Tuple[Dict[str, str], Any]]:
+        """
+        :param args:
+        :param return_headers:  if set return a tuple (header, body)
+        :param kwargs:
+        :return: body or (header, body)
+        """
+        headers, data, result = self.request(*args, **kwargs)
+        if return_headers:
+            return headers, data
+        return data
 
     def _factory_args(self):
         return {"auth": self.req.auth, "headers": {"user-agent": f"aiopenapi3/{__version__}"}}
 
-    def request(
-        self, data=None, parameters=None, return_headers: bool = False
-    ) -> Union[Any, Tuple[Dict[str, Any], Any]]:
+    def request(self, data=None, parameters=None) -> Tuple[Dict[str, Any], Any, httpx.Response]:
         """
         Sends an HTTP request as described by this Path
 
@@ -62,37 +71,54 @@ class RequestBase:
         :type data: any, should match content/type
         :param parameters: The parameters used to create the path
         :type parameters: dict{str: str}
-        :param return_headers: if set return a tuple (header, body)
-        :return: body or (header, body)
+        :return: headers, data, response
         """
         self._prepare(data, parameters)
         with closing(self.api._session_factory(**self._factory_args())) as session:
             req = self._build_req(session)
-            result = session.send(req)
+            try:
+                result = session.send(req)
+            except Exception:
+                raise RequestError(self.operation, req, data, parameters)
         headers, data = self._process(result)
-        if return_headers is True:
-            return headers, data
-        else:
-            return data
+
+        return headers, data, result
+
+    @property
+    @abc.abstractmethod
+    def data(self) -> "SchemaBase":
+        """
+        :return: the Schema for the body
+        """
+        pass
+
+    @property
+    @abc.abstractmethod
+    def parameters(self) -> List["ParameterBase"]:
+        """
+        :return: list of parameters
+        """
+        pass
 
 
 class AsyncRequestBase(RequestBase):
-    async def __call__(self, *args, **kwargs):
-        return await self.request(*args, **kwargs)
+    async def __call__(self, *args, return_headers: bool = False, **kwargs):
+        headers, data, result = await self.request(*args, **kwargs)
+        if return_headers:
+            return headers, data
+        return data
 
-    async def request(
-        self, data=None, parameters=None, return_headers: bool = False
-    ) -> Union[Any, Tuple[Dict[str, Any], Any]]:
+    async def request(self, data=None, parameters=None) -> Tuple[Dict[str, Any], Any, httpx.Response]:
         self._prepare(data, parameters)
         async with aclosing(self.api._session_factory(**self._factory_args())) as session:
             req = self._build_req(session)
-            result = await session.send(req)
+            try:
+                result = await session.send(req)
+            except Exception as e:
+                raise RequestError(self.operation, req, data, parameters)
 
         headers, data = self._process(result)
-        if return_headers is True:
-            return headers, data
-        else:
-            return data
+        return headers, data, result
 
 
 class OperationIndex:
