@@ -1,3 +1,4 @@
+import enum
 from typing import Union, Optional, Dict, Any
 
 from pydantic import Field
@@ -8,6 +9,8 @@ from .example import Example
 from .general import Reference
 from .schemas import Schema
 
+from ..v30.parameter import _ParameterCodec
+
 
 class ParameterBase(ObjectExtended):
     """
@@ -31,18 +34,26 @@ class ParameterBase(ObjectExtended):
     content: Optional[Dict[str, "MediaType"]]
 
 
-class Parameter(ParameterBase):
+class _In(str, enum.Enum):
+    query = "query"
+    header = "header"
+    path = "path"
+    cookie = "cookie"
+
+
+class Parameter(ParameterBase, _ParameterCodec):
     name: str = Field(required=True)
-    in_: str = Field(required=True, alias="in")  # TODO must be one of ["query","header","path","cookie"]
+    in_: _In = Field(required=True, alias="in")
 
 
-class Header(ParameterBase):
+class Header(ParameterBase, _ParameterCodec):
     """
 
     .. _HeaderObject: https://github.com/OAI/OpenAPI-Specification/blob/main/versions/3.1.0.md#headerObject
     """
 
-    pass
+    def _codec(self):
+        return "simple", False
 
 
 from .media import MediaType
